/**
 *  OdetteJ - Open Source Odette FTP implementation in Java(TM).
 *  Copyright (C) 2005 Fossil E-Commerce, http://www.fossilec.com
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  ------------------------------------------------------------------------
 *  $Id: AbstractState.java,v 1.9 2005/09/12 19:20:28 rmarins Exp $
 *  ------------------------------------------------------------------------
 */

package org.fossilec.odettej;

import org.fossilec.odettej.service.CommandExchangeBuffer;
import org.fossilec.odettej.service.CommandIdentifier;
import org.fossilec.odettej.service.EndSessionReason;
import org.fossilec.odettej.service.OdetteExchangeBuffer;
import org.fossilec.odettej.service.Service;
import org.fossilec.odettej.service.TransferMode;
import org.fossilec.odettej.transport.Transport;

abstract class AbstractState implements OdetteFTPEntity {

	private Session context;

	private PreferencesUtil prefs;

	private boolean useStrictFormat;

	protected AbstractState(Session context) {
		super();

		this.context = context;
		prefs = new PreferencesUtil(context.getProperties());

		useStrictFormat = prefs.isUsingStrictFormat();
	}

	protected abstract void listen() throws OdetteFTPException;

	public Session getContext() {
		return context;
	}

	public PreferencesUtil getPreferences() {
		return prefs;
	}

	public Service getService() throws OdetteFTPException {
		return context.getService();
	}

	public void resetCredits() {
		String window = context.getProperty("odette.window-size");
		setCredits(Integer.parseInt(window));
	}

	public void consumeCredit() {
		int credits = getCredits();
		credits--;

		setCredits(credits);
	}

	protected abstract void setCredits(int windowSize);

	public abstract int getCredits();

	public boolean hasCredits() {
		return (getCredits() > 0);
	}

	protected OdetteExchangeBuffer receive() throws OdetteFTPException {
		OdetteExchangeBuffer exchangeBuffer = getService().receive(
				useStrictFormat);

		if (exchangeBuffer.getIdentifier() == CommandIdentifier.ESID) {

			CommandExchangeBuffer esid = (CommandExchangeBuffer) exchangeBuffer;

			String code = esid.getParameter("ESIDREAS");
			EndSessionReason reason = EndSessionReason.parse(code);

			if (reason != EndSessionReason.NORMAL_TERMINATION) {
				abnormalRelease(reason);
			}
		}

		return exchangeBuffer;
	}

	protected void send(OdetteExchangeBuffer oeb) throws OdetteFTPException {
		Service service = getService();
		service.send(oeb);
	}

	protected void disconnect() throws OdetteFTPException {
		context.disconnect();
	}

	public boolean isConnected() throws OdetteFTPException {
		Service service = getService();
		Transport transport = service.getTransport();

		return transport.isConnected();
	}

	public void release() throws OdetteFTPException {
		protocolRelease(EndSessionReason.NORMAL_TERMINATION);
	}

	protected void abnormalRelease(EndSessionReason error)
			throws OdetteFTPException {

		protocolRelease(error);
		endSessionError(error, "abnormal release");
	}

	protected void abnormalRelease(EndSessionReason error, String msg)
			throws OdetteFTPException {

		protocolRelease(error);
		endSessionError(error, msg);
	}

	public void abort(EndSessionReason error) throws OdetteFTPException {
		protocolRelease(error);
	}

	private void protocolRelease(EndSessionReason reason)
			throws OdetteFTPException {

		Service service = getService();

		CommandExchangeBuffer esid;

		if (reason == null) {
			throw new IllegalArgumentException("Reason must not be null");
		}

		/*
		 * Sending End Session command.
		 */
		esid = CommandExchangeBuffer.endSession(reason);

		send(esid);

		try {
			Thread.sleep(500);
		} catch (Exception e) {
			e.printStackTrace();
		}

		service.disconnect();

		context.changeState(IdleState.class);
	}

	private static void endSessionError(EndSessionReason error, String msg)
			throws OdetteFTPException {
		throw new EndSessionException(error, msg);
	}

	protected void requestAuthentication(String ssidcode, String ssidpswd,
			String remoteAddress) throws OdetteFTPException {

		Session context = getContext();

		PeerAuthenticator auth = context.getAuthenticator();

		if (auth != null) {
			auth.authenticate(ssidcode, ssidpswd, remoteAddress);
		}
	}

	protected static boolean windowSizeViolation(int size, Session session) {
		return ((size <= 0) || (size > session.getMaxWindow()));
	}

	protected static boolean bufferSizeViolation(int size, Session session) {
		return ((size < OdetteExchangeBuffer.MIN_OEB_LENGTH) || (size > session
				.getMaxBuffer()));
	}

	protected static boolean valueOfYesNo(String parameter) {
		return (parameter.equals("Y") ? true : false);
	}

	protected static void checkStartSession(Session context, int sdeb,
			int window, boolean compression, TransferMode mode, boolean restart)
			throws OdetteFTPException {

		/* check exchange buffer size */
		if (bufferSizeViolation(sdeb, context)) {
			endSessionError(EndSessionReason.INVALID_COMMAND_DATA,
					"Invalid exchange buffer size is being used: " + sdeb);
		}

		/* check window size */
		if (windowSizeViolation(window, context)) {
			endSessionError(EndSessionReason.INVALID_COMMAND_DATA,
					"Invalid window size is being used: " + window);
		}

		/* check compression capability */
		if ((!context.isCompressionCapable()) && (compression)) {
			endSessionError(EndSessionReason.INCOMPATIBLE_MODE,
					"Implementation doesn't support compression");
		}

		/* check transfer mode capability */
		TransferMode capmode = context.getCapableTransferMode();
		if ((capmode != TransferMode.BOTH) && (mode != capmode)) {
			endSessionError(EndSessionReason.INCOMPATIBLE_MODE,
					"Implementation doesn't support this mode: " + mode);
		}

		/* check restart capability */
		if ((!context.isRestartCapable()) && (restart)) {
			endSessionError(EndSessionReason.INCOMPATIBLE_MODE,
					"Implementation doesn't support restart");
		}
	}

}
