/**
 *  OdetteJ - Open Source Odette FTP implementation in Java(TM).
 *  Copyright (C) 2005 Fossil E-Commerce, http://www.fossilec.com
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  ------------------------------------------------------------------------
 *  $Id: ListenerState.java,v 1.14 2005/09/10 23:58:11 rmarins Exp $
 *  ------------------------------------------------------------------------
 */

package org.fossilec.odettej;

import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.WeakHashMap;

import org.fossilec.odettej.VirtualFile.MappingMode;
import org.fossilec.odettej.service.AnswerReason;
import org.fossilec.odettej.service.CommandExchangeBuffer;
import org.fossilec.odettej.service.CommandIdentifier;
import org.fossilec.odettej.service.DataExchangeBuffer;
import org.fossilec.odettej.service.EndSessionReason;
import org.fossilec.odettej.service.OdetteExchangeBuffer;
import org.fossilec.odettej.service.RecordFormat;
import org.fossilec.odettej.service.TransferMode;

class ListenerState extends AbstractState {

	/**
	 * Estimated receiving File Size Blocks
	 */
	private DataExchangeBuffer data;

	private List eerpQueue = new ArrayList();

	private String transferDestination;

	private String transferOriginator;

	private String transferUserData;

	private String transferReserved;

	private String transferDatasetName;

	public ListenerState(Session context) {
		super(context);
	}

	protected void listen() throws OdetteFTPException {

		OdetteExchangeBuffer exchangeBuffer;

		/* Idle Listener state */
		while ((exchangeBuffer = receive()) != null) {

			CommandIdentifier ci = exchangeBuffer.getIdentifier();

			if (ci == CommandIdentifier.DATA) {
				answerData(exchangeBuffer);

			} else if (ci == CommandIdentifier.EERP) {
				answerEndToEndResponse(exchangeBuffer);

			} else if (ci == CommandIdentifier.SFID) {
				answerStartFile(exchangeBuffer);

			} else if (ci == CommandIdentifier.EFID) {
				answerCloseFile(exchangeBuffer);

			} else if (ci == CommandIdentifier.CD) {
				changeDirection();
				break;

			} else if (ci == CommandIdentifier.ESID) {
				answerEndSession(exchangeBuffer);
				break;

			} else {
				abnormalRelease(EndSessionReason.PROTOCOL_VIOLATION,
						"Invalid command to the current state: "
								+ ci.toString());
			}
		}
	}

	public void startSession() throws OdetteFTPException {

		Session context = getContext();
		PreferencesUtil prefs = getPreferences();

		TransferMode mode = prefs.getTransferMode();
		boolean restart = prefs.isRestartSupported();

		/* Beging Start Session Phase */
		sendReadyMessage();

		negotiateStartSession(mode, restart);

		/*
		 * As the initiator will always be the Speaker, when is specified the
		 * receiver only just change direction after establishing a session.
		 */
		if (prefs.getTransferMode() == TransferMode.RECEIVER_ONLY) {
			context.state().changeDirection();
		}

	}

	private void sendReadyMessage() throws OdetteFTPException {

		CommandExchangeBuffer ssrm = CommandExchangeBuffer.readyMessage();

		send(ssrm);

	}

	private void answerData(OdetteExchangeBuffer exchangeBuffer)
			throws OdetteFTPException {

		this.data = (DataExchangeBuffer) exchangeBuffer;
		dataRegime();
	}

	private void answerStartFile(OdetteExchangeBuffer exchangeBuffer)
			throws OdetteFTPException {

		Session session = getContext();
		ListenerTemplate listenerSupport = session.getListenerSupport();

		/*
		 * Extract command parameters
		 */
		CommandExchangeBuffer sfid = (CommandExchangeBuffer) exchangeBuffer;

		String datasetName = sfid.getParameter("SFIDDSN");
		String reserved = sfid.getParameter("SFIDRSV1");
		String fileDate = sfid.getParameter("SFIDDATE");
		String fileTime = sfid.getParameter("SFIDTIME");
		String userData = sfid.getParameter("SFIDUSER");
		String destination = sfid.getParameter("SFIDDEST");
		String originator = sfid.getParameter("SFIDORIG");
		RecordFormat format = (RecordFormat) RecordFormat.parse(sfid
				.getParameter("SFIDFMT"));
		int maxRecordSize = Integer.parseInt(sfid.getParameter("SFIDLRECL"));
		int fileSizeBlocks = Integer.parseInt(sfid.getParameter("SFIDFSIZ"));
		long restartPosition = Long.parseLong(sfid.getParameter("SFIDREST"));

		Date fileDateTime = StringUtil.parseCommandDateTime(fileDate, fileTime);

		/*
		 * Cases of Start File Negative Answer
		 */
		if (listenerSupport == null) {
			startFileNegativeAnswer(AnswerReason.UNSPECIFIED, true);
			return;
		} else {
			AnswerReason answer = listenerSupport.requestStartFile(datasetName,
					fileDateTime, destination, originator, format,
					maxRecordSize, fileSizeBlocks, restartPosition, userData,
					reserved);

			if (answer != null) {
				startFileNegativeAnswer(answer, true);
				return;
			}
		}

		/*
		 * Start File
		 */

		/*
		 * Set class variables used by closeFile() method later.
		 */
		transferDatasetName = datasetName;

		/* Pick up Virtual File from the listener support provider */
		VirtualFile file = null;

		try {
			file = listenerSupport.beginStartFile(datasetName, fileDateTime,
					destination, originator, format, maxRecordSize,
					fileSizeBlocks, restartPosition, userData, reserved);
		} catch (FileTransferException e) {
			CommandExchangeBuffer efna = CommandExchangeBuffer
					.endFileNegativeAnswer(e.getReason());
			send(efna);

			return;
		}

		startFile(file, destination, originator, userData, reserved);

	}

	private void answerCloseFile(OdetteExchangeBuffer exchangeBuffer)
			throws OdetteFTPException {

		CommandExchangeBuffer efid = (CommandExchangeBuffer) exchangeBuffer;

		int recordCount = Integer.parseInt(efid.getParameter("EFIDRCNT"));
		long unitCount = Long.parseLong(efid.getParameter("EFIDUCNT"));

		closeFile(recordCount, unitCount);

		this.data = null;
	}

	private void answerEndToEndResponse(OdetteExchangeBuffer exchangeBuffer)
			throws OdetteFTPException {

		CommandExchangeBuffer eerp = (CommandExchangeBuffer) exchangeBuffer;

		String datasetName = eerp.getParameter("EERPDSN");
		String reserved = eerp.getParameter("EERPRSV1");
		String fileDate = eerp.getParameter("EERPDATE");
		String fileTime = eerp.getParameter("EERPTIME");
		String userData = eerp.getParameter("EERPUSER");
		String destination = eerp.getParameter("EERPDEST");
		String originator = eerp.getParameter("EERPORIG");

		Date fileDateTime = StringUtil.parseCommandDateTime(fileDate, fileTime);

		endToEndResponse(datasetName, fileDateTime, destination, originator,
				userData, reserved);
	}

	/**
	 * @param exchangeBuffer
	 * @throws OdetteFTPException
	 */
	private void answerEndSession(OdetteExchangeBuffer exchangeBuffer)
			throws OdetteFTPException {

		disconnect();
	}

	public void startFile(VirtualFile file, String destination,
			String originator, String userData, String reserved)
			throws OdetteFTPException {

		Session context = getContext();

		if ((file == null) || (file.getMode() != MappingMode.OUTPUT_MAPPING)) {
			abnormalRelease(EndSessionReason.INCOMPATIBLE_MODE,
					"Unhandled start file indication");
		}

		/*
		 * Start File Positive Answer
		 */

		int answerCount = file.getOffset();
		startFilePositiveAnswer(answerCount);

		/* the SFPA command acts as a implicit Set Credit command */
		resetCredits();

		context.setFile(file);

		/*
		 * Set class variables used by closeFile() method later.
		 */
		transferDestination = destination;
		transferOriginator = originator;
		transferUserData = userData;
		transferReserved = reserved;

	}

	private void startFilePositiveAnswer(int answerCount)
			throws OdetteFTPException {

		CommandExchangeBuffer sfpa = CommandExchangeBuffer
				.startFilePositiveAnswer(answerCount);

		send(sfpa);
	}

	private void startFileNegativeAnswer(AnswerReason reason, boolean retry)
			throws OdetteFTPException {

		CommandExchangeBuffer sfna = CommandExchangeBuffer
				.startFileNegativeAnswer(reason, retry);
		send(sfna);
	}

	public void dataRegime() throws OdetteFTPException {

		Session context = getContext();
		VirtualFile file = context.getFile();

		PreferencesUtil prefs = getPreferences();
		boolean compression = prefs.isCompressionSupported();
		RecordFormat recordFormat = file.getFormat();

		/* write the received buffer to the Virtual File */
		ByteBuffer flattenData = data.scrape(compression, recordFormat);
		file.writeData(flattenData);

		/* credits consuming and resetting handling */
		consumeCredit();

		if (!hasCredits()) {
			setCredit();
		}
	}

	private void setCredit() throws OdetteFTPException {
		CommandExchangeBuffer cdt = CommandExchangeBuffer.setCredit("");
		send(cdt);

		resetCredits();
	}

	public void closeFile(int recordCount, long unitCount)
			throws OdetteFTPException {

		Session context = getContext();

		VirtualFile file = context.getFile();

		RecordFormat format = file.getFormat();
		long fileByteCount = file.getUnitCount();

		String datasetName = file.getDatasetName();
		Date fileDateTime = file.getDateTime();

		/* close file anyway */
		file.close();

		/*
		 * Record Count comparison:
		 * 
		 * TEXTFILE and UNSTRUCTURED formats must have EFIDRCNT equal to zero.
		 * FIXED and VARIABLE formats must have EFIDRCNT exactly equals.
		 */
		boolean invalidTextUnstructured = (recordCount != 0)
				&& (format == RecordFormat.TEXTFILE)
				&& (format == RecordFormat.UNSTRUCTURED);

		boolean invalidFixedVariable = (recordCount != file.getRecordCount())
				&& (format == RecordFormat.FIXED)
				&& (format == RecordFormat.VARIABLE);

		if (invalidTextUnstructured || invalidFixedVariable) {
			endFileNegativeAnswer(AnswerReason.INVALID_RECORD_COUNT);
			return;
		}

		/*
		 * File Size comparison
		 */
		if (unitCount != fileByteCount) {
			endFileNegativeAnswer(AnswerReason.INVALID_BYTE_COUNT);
			return;
		}

		/*
		 * In case of delivery acknowledgment after receiving, is ready to send
		 * back an EERP command.
		 */
		ListenerTemplate listenerSupport = context.getListenerSupport();

		boolean sendEERP = listenerSupport.hasDeliveredFile(datasetName,
				fileDateTime, transferDestination, transferOriginator);

		/*
		 * Send End File Positive Answer
		 */
		endFilePositiveAnswer(sendEERP);

		/*
		 * End To End Response are queued to be sent later when the Change
		 * Direction command is received.
		 */
		if (sendEERP) {

			queueEndToEndResponse(transferDatasetName, transferReserved,
					fileDateTime, transferUserData, transferDestination,
					transferOriginator);
		}

		/* destroy class variables */
		transferDestination = null;
		transferOriginator = null;
		transferUserData = null;
		transferReserved = null;

	}

	/**
	 * @param datasetName
	 * @param reserved
	 * @param fileDateTime
	 * @param userData
	 * @param destination
	 * @param originator
	 */
	private void queueEndToEndResponse(String datasetName, String reserved,
			Date fileDateTime, String userData, String destination,
			String originator) {

		Map eerp = new WeakHashMap(6);

		eerp.put("datasetName", datasetName);
		eerp.put("fileDateTime", fileDateTime);
		eerp.put("userData", userData);
		eerp.put("destination", destination);
		eerp.put("originator", originator);
		eerp.put("reserved", reserved);

		eerpQueue.add(eerp);

	}

	/**
	 * @param eerp
	 * @throws OdetteFTPException
	 */
	private void sendQueuedEERP(Map eerp) throws OdetteFTPException {

		Session context = getContext();

		context.state().endToEndResponse((String) eerp.get("datasetName"),
				(Date) eerp.get("fileDateTime"),
				(String) eerp.get("destination"),
				(String) eerp.get("originator"), (String) eerp.get("userData"),
				(String) eerp.get("reserved"));

	}

	private void endFileNegativeAnswer(AnswerReason reason)
			throws OdetteFTPException {

		CommandExchangeBuffer efna = CommandExchangeBuffer
				.endFileNegativeAnswer(reason);

		send(efna);
	}

	private void endFilePositiveAnswer(boolean changeDirection)
			throws OdetteFTPException {

		CommandExchangeBuffer efpa = CommandExchangeBuffer
				.endFilePositiveAnswer(changeDirection);

		send(efpa);
	}

	public void changeDirection() throws OdetteFTPException {

		Session context = getContext();
		PreferencesUtil prefs = getPreferences();

		/* become Speaker */
		context.changeState(SpeakerState.class);

		if (prefs.getTransferMode() == TransferMode.RECEIVER_ONLY) {

			/*
			 * A Change Direction request in the Receiver Only mode means that
			 * the another peer has no more files to be sent, so the session
			 * must be finished.
			 */

			context.state().release();

		} else {

			/*
			 * Send back lasting EERP commands queued and keep on Sender Mode.
			 */

			Iterator it = eerpQueue.iterator();
			while (it.hasNext()) {

				Map eerp = (Map) it.next();

				sendQueuedEERP(eerp);

				it.remove();
			}

		}

	}

	public void endToEndResponse(String datasetName, Date fileDateTime,
			String destination, String originator, String userData,
			String reserved) throws OdetteFTPException {

		Session session = getContext();

		/* delegate delivery acknowledge to the listener support object */
		ListenerTemplate support = session.getListenerSupport();
		if (support != null) {
			support.fileDelivered(datasetName, fileDateTime, destination,
					originator, userData, reserved);
		}

		/**
		 * Ready to Receive command
		 */

		CommandExchangeBuffer rtr;
		rtr = CommandExchangeBuffer.readyToReceive();

		send(rtr);
	}

	private void negotiateStartSession(TransferMode mode, boolean restart)
			throws OdetteFTPException {

		Session context = getContext();
		PreferencesUtil prefs = getPreferences();

		/*
		 * Retrieve and check session parameters.
		 */
		int sdeb = prefs.getBufferSize();
		boolean compression = prefs.isCompressionSupported();
		boolean specialLogic = prefs.hasSpecialLogic();
		int window = prefs.getWindowSize();

		checkStartSession(context, sdeb, window, compression, mode, restart);

		/*
		 * Receiving Start Session command.
		 */
		CommandExchangeBuffer receivedCmd = (CommandExchangeBuffer) receive();

		if (receivedCmd.getIdentifier() == CommandIdentifier.SSID) {

			/* check values negotiated on session and do set up */
			sessionSetup(sdeb, mode, compression, restart, specialLogic,
					window, receivedCmd);

		} else {
			abnormalRelease(EndSessionReason.PROTOCOL_VIOLATION,
					"SSID commands were expected");
		}

	}

	private void sessionSetup(int sdeb, TransferMode mode, boolean compression,
			boolean restart, boolean specialLogic, int window,
			CommandExchangeBuffer ssid) throws OdetteFTPException {

		PreferencesUtil prefs = getPreferences();

		String ssidcode = ssid.getParameter("SSIDCODE");
		String ssidpswd = ssid.getParameter("SSIDPSWD");
		int ssidsdeb = Integer.parseInt(ssid.getParameter("SSIDSDEB"));
		TransferMode ssidsr = (TransferMode) TransferMode.parse(ssid
				.getParameter("SSIDSR"));
		boolean ssidcmpr = valueOfYesNo(ssid.getParameter("SSIDCMPR"));
		boolean ssidrest = valueOfYesNo(ssid.getParameter("SSIDREST"));
		boolean ssidspec = valueOfYesNo(ssid.getParameter("SSIDSPEC"));
		int ssidcred = Integer.parseInt(ssid.getParameter("SSIDCRED"));

		/* check returning credentials */
		requestAuthentication(ssidcode, ssidpswd, prefs.getCallingAddress());

		/*
		 * Session Negotiation and Configure
		 */

		if ((!compression) && (ssidcmpr)) {
			ssidcmpr = false;
		}

		if ((!restart) && (ssidrest)) {
			ssidrest = false;
		}

		TransferMode ssidsrReversed = ssidsr.getReversed();
		if ((ssidsrReversed != TransferMode.BOTH) && (mode != ssidsrReversed)) {
			abnormalRelease(EndSessionReason.INCOMPATIBLE_MODE,
					"Invalid mode must be: " + mode);
		} else if (ssidsrReversed == TransferMode.BOTH) {
			ssidsrReversed = mode;
		}

		if (specialLogic != ssidspec) {
			abnormalRelease(EndSessionReason.INCOMPATIBLE_MODE,
					"wrong special logic indicator specified by peer");
		}

		if (ssidsdeb > sdeb) {
			ssidsdeb = sdeb;
		}

		if (ssidcred > window) {
			ssidcred = window;
		}

		prefs.setCompressionSupport(ssidcmpr);
		prefs.setRestartSupport(ssidrest);
		prefs.setTransferMode(ssidsrReversed);
		prefs.setSpecialLogic(ssidspec);
		prefs.setBufferSize(ssidsdeb);
		prefs.setWindowSize(ssidcred);

		/*
		 * Send Start Session ID
		 */

		/* so far to retrive usercode and password to send back SSID */
		String userCode = prefs.getLocalUser();
		String password = prefs.getLocalPassword();

		String reserved = prefs.getReservedParam();
		String userData = prefs.getUserData();

		CommandExchangeBuffer goingCmd = CommandExchangeBuffer.startSession(
				userCode, password, ssidsdeb, ssidsrReversed, ssidcmpr,
				ssidrest, ssidspec, ssidcred, reserved, userData);

		send(goingCmd);
	}

	protected void setCredits(int windowSize) {
		Session context = getContext();
		context.setProperty("odette.credits.listener", String
				.valueOf(windowSize));
	}

	public int getCredits() {
		Session context = getContext();
		String credits = context.getProperty("odette.credits.listener");

		return Integer.parseInt(credits);
	}

}
