/**
 *  OdetteJ - Open Source Odette FTP implementation in Java(TM).
 *  Copyright (C) 2005 Fossil E-Commerce, http://www.fossilec.com
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  ------------------------------------------------------------------------
 *  $Id: ConnectionMonitor.java,v 1.1 2005/09/10 23:58:45 rmarins Exp $
 *  ------------------------------------------------------------------------
 */

package org.fossilec.odettej.monitor;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.InetSocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.SocketChannel;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Iterator;
import java.util.logging.Logger;

/**
 * 
 * @author Rafael Marins
 */
public class ConnectionMonitor implements Runnable {

	private SocketChannel initiatorChannel;

	private SocketChannel responderChannel;

	private String remoteHost;

	private int remotePort;

	private FileChannel logfile;

	private Logger logger;

	private ByteBuffer buffer;

	private int loggingOctets;

	private boolean stopped = false;

	public ConnectionMonitor(SocketChannel client, FileChannel logfile,
			String host, int port, int blockSize, int readOctets) {

		this.initiatorChannel = client;
		this.logfile = logfile;
		this.remoteHost = host;
		this.remotePort = port;

		this.logger = Logger.global;
		this.buffer = ByteBuffer.allocateDirect(blockSize);
		this.loggingOctets = readOctets;
	}

	public void run() {

		Selector selector;

		try {
			selector = Selector.open();

			openRemoteConnection(selector);

			while (!stopped) {

				int count = selector.select();

				if (count == 0) {
					continue;
				}

				Iterator it = selector.selectedKeys().iterator();

				while (it.hasNext()) {

					SelectionKey key = (SelectionKey) it.next();

					// new read operation
					if (key.isReadable()) {
						SocketChannel source = (SocketChannel) key.channel();
						doRead(source, key.attachment());
					}

					// new connection operation
					if (key.isConnectable()) {
						SocketChannel source = (SocketChannel) key.channel();
						doConnect(source, selector);
					}

					// remove key from selected set; it's been handled
					it.remove();
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
		} finally {
			doClose();

			Thread.currentThread().interrupt();
		}

	}

	public void stop() {
		this.stopped = true;
	}

	private void doClose() {

		try {
			initiatorChannel.close();
		} catch (IOException e) {
			e.printStackTrace();
		}

		try {
			responderChannel.close();
		} catch (IOException e) {
			e.printStackTrace();
		}

		try {
			logfile.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	private void openRemoteConnection(Selector selector) throws Exception {

		responderChannel = SocketChannel.open();

		SelectionKey key = TCPMonitor.registerChannel(selector,
				responderChannel, SelectionKey.OP_CONNECT);

		logger.info("Redirecting connection to destination host...");
		responderChannel.connect(new InetSocketAddress(remoteHost, remotePort));

		if (!responderChannel.isConnectionPending()) {
			key.cancel();

			beginReadingChannels(selector);
		}
	}

	private void beginReadingChannels(Selector selector) throws Exception {

		TCPMonitor.registerChannel(selector, responderChannel,
				SelectionKey.OP_READ, "Responder");

		TCPMonitor.registerChannel(selector, initiatorChannel,
				SelectionKey.OP_READ, "Initiator");

	}

	protected void finalize() throws Throwable {
		doClose();
		super.finalize();
	}

	public void doAccept(Selector selector) throws Exception {

		responderChannel = SocketChannel.open();

		SelectionKey key = TCPMonitor.registerChannel(selector,
				responderChannel, SelectionKey.OP_CONNECT);

		logger.info("Redirecting connection to destination host...");
		responderChannel.connect(new InetSocketAddress(remoteHost, remotePort));

		if (!responderChannel.isConnectionPending()) {
			key.cancel();

			TCPMonitor.registerChannel(selector, responderChannel,
					SelectionKey.OP_READ, "Responder");

			TCPMonitor.registerChannel(selector, initiatorChannel,
					SelectionKey.OP_READ, "Initiator");
		}
	}

	private void doConnect(SocketChannel responder, Selector selector)
			throws Exception {

		if (!responder.finishConnect()) {
			logger.info("Connection is pending...");
			return;
		}

		logger.info("Connection established to "
				+ responder.socket().getRemoteSocketAddress());

		beginReadingChannels(selector);
	}

	private synchronized void doRead(SocketChannel source, Object attachment)
			throws Exception {

		String mode = (String) attachment;
		SocketChannel peer = null;

		if (source.equals(initiatorChannel)) {
			peer = responderChannel;
		} else if (source.equals(responderChannel)) {
			peer = initiatorChannel;
		} else {
			return;
		}

		if ((!source.isConnected()) || (!peer.isConnected())) {
			stop();
		}

		int read = 0;

		try {
			while (source.read(buffer) > 0) {

				if (read == 0) {
					saveLog("");
				}

				buffer.flip();
				read += buffer.limit();

				if (read <= loggingOctets) {
					doPrint(buffer, mode);
				}

				buffer.rewind();
				peer.write(buffer);
				buffer.clear();
			}
		} catch (IOException e) {
			stop();
		}
	}

	private void doPrint(ByteBuffer data, String mode) {
		String msg = mode + " - [" + parseOctetsSequence(data) + "] - {"
				+ parseCharSequence(data) + "}";
		saveLog(msg);
		logger.fine(msg);
	}

	private void saveLog(String msg) {

		SimpleDateFormat df = new SimpleDateFormat("dd/HH/yyyy HH:mm:ss,SSS");
		String timestamp = df.format(Calendar.getInstance().getTime());

		String octets = timestamp + " - " + msg + "\n";

		try {
			logfile.write(ByteBuffer.wrap(octets.getBytes("US-ASCII")));
		} catch (IOException e) {
			e.printStackTrace();
		}

	}

	private static String parseCharSequence(ByteBuffer data) {

		StringBuffer sequence = new StringBuffer();

		while (data.hasRemaining()) {
			byte octet = (byte) getUnsignedByte(data);
			try {
				if ((octet < 30) || (octet > 128)) {
					sequence.append(".");
				} else {
					sequence
							.append(new String(new byte[] { octet }, "US-ASCII"));
				}
			} catch (UnsupportedEncodingException e) {
				e.printStackTrace();
			}
		}

		data.rewind();

		return sequence.toString();
	}

	private static String parseOctetsSequence(ByteBuffer data) {

		int count = 0;

		StringBuffer sequence = new StringBuffer();

		while (data.hasRemaining()) {
			count++;

			short octet = getUnsignedByte(data);

			if (octet < 16)
				sequence.append("0" + Integer.toHexString(octet));
			else
				sequence.append(Integer.toHexString(octet));

			if ((count % 4 == 0) && data.hasRemaining())
				sequence.append(" ");
		}

		data.rewind();

		return sequence.toString();
	}

	private static short getUnsignedByte(ByteBuffer bb) {
		return ((short) (bb.get() & 0xff));
	}

}
