/**
 *  OdetteJ - Open Source Odette FTP implementation in Java(TM).
 *  Copyright (C) 2005 Fossil E-Commerce, http://www.fossilec.com
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  ------------------------------------------------------------------------
 *  $Id: TCPMonitor.java,v 1.1 2005/09/10 23:58:45 rmarins Exp $
 *  ------------------------------------------------------------------------
 */

package org.fossilec.odettej.monitor;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.InetSocketAddress;
import java.nio.channels.FileChannel;
import java.nio.channels.SelectableChannel;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.ServerSocketChannel;
import java.nio.channels.SocketChannel;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Iterator;
import java.util.logging.Logger;

/**
 * 
 * @author Rafael Marins
 */
public class TCPMonitor {

	private int localPort;

	private String remoteHost;

	private int remotePort;

	private String logPrefix;

	private int blockSize;

	private int bytesToRead;

	private Logger logger;

	public TCPMonitor(int listening, String host, int port, String logPrefix,
			int blockSize, int bytesToRead) {

		super();

		this.localPort = listening;
		this.remoteHost = host;
		this.remotePort = port;
		this.logPrefix = logPrefix;
		this.blockSize = blockSize;
		this.bytesToRead = bytesToRead;

		this.logger = Logger.global;
	}

	public void init() throws Exception {

		logger.info("Starting TCPMonitor on port " + localPort + "...");

		ServerSocketChannel ssc = ServerSocketChannel.open();
		ssc.socket().bind(new InetSocketAddress(localPort));

		Selector selector = Selector.open();

		registerChannel(selector, ssc, SelectionKey.OP_ACCEPT);

		while (true) {

			int count = selector.select();

			if (count == 0) {
				continue;
			}

			Iterator it = selector.selectedKeys().iterator();

			while (it.hasNext()) {

				SelectionKey key = (SelectionKey) it.next();

				// new incoming connection
				if (key.isAcceptable()) {

					ServerSocketChannel server = (ServerSocketChannel) key
							.channel();

					doAccept(server);
				}

				// remove key from selected set; it's been handled
				it.remove();
			}
		}
	}

	private void doAccept(ServerSocketChannel server) throws Exception {

		// accept incoming connection and register in selector
		SocketChannel client = server.accept();

		logger.info("Accepting incoming connection from "
				+ client.socket().getRemoteSocketAddress());

		FileChannel logfile = openLogFile(logPrefix);
		ConnectionMonitor connection = new ConnectionMonitor(client, logfile,
				remoteHost, remotePort, blockSize, bytesToRead);

		Thread process = new Thread(connection);
		process.start();

	}

	public static SelectionKey registerChannel(Selector selector,
			SelectableChannel channel, int ops, Object attachment)
			throws IOException {

		if (channel == null) {
			return null;
		}

		channel.configureBlocking(false);

		return channel.register(selector, ops, attachment);

	}

	public static SelectionKey registerChannel(Selector selector,
			SelectableChannel channel, int ops) throws IOException {
		return registerChannel(selector, channel, ops, null);
	}

	private static FileChannel openLogFile(String prefix) {

		SimpleDateFormat df = new SimpleDateFormat("yyMMddHHmmss'-'SSS");
		String timestamp = df.format(Calendar.getInstance().getTime());

		FileChannel logfile = null;

		try {
			File file = new File(prefix + "." + timestamp);
			FileOutputStream os = new FileOutputStream(file, true);

			logfile = os.getChannel();
		} catch (IOException e) {
			e.printStackTrace();
		}

		return logfile;
	}

	public static void main(String[] args) throws Exception {

		int localPort;
		int remotePort;
		int blockSize = 16;
		int bytesToRead = 128;

		String remoteHost;
		String logPrefix;

		if ((args.length < 4) || (args.length > 5)) {
			System.out.println("Usage: TCPMonitor <localPort> <remoteHost> "
					+ "<remotePort> <loggingPrefix> [<blockSize> "
					+ "[<bytesToRead>]]");
			System.exit(-1);
		}

		localPort = Integer.parseInt(args[0]);
		remoteHost = args[1];
		remotePort = Integer.parseInt(args[2]);
		logPrefix = args[3];

		if (args.length >= 5) {
			blockSize = Integer.parseInt(args[4]);
		}

		if (args.length == 6) {
			blockSize = Integer.parseInt(args[5]);
		}

		TCPMonitor monitor = new TCPMonitor(localPort, remoteHost, remotePort,
				logPrefix, blockSize, bytesToRead);

		monitor.init();

	}
}
