/**
 *  OdetteJ - Open Source Odette FTP implementation in Java(TM).
 *  Copyright (C) 2005 Fossil E-Commerce, http://www.fossilec.com
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  ------------------------------------------------------------------------
 *  $Id: DataExchangeBuffer.java,v 1.7 2005/08/15 16:43:36 rmarins Exp $
 *  ------------------------------------------------------------------------
 */

package org.fossilec.odettej.service;

import java.nio.ByteBuffer;

import org.fossilec.odettej.OdetteFTPException;

/**
 * DataExchange is an type of Odette FTP Exchange Buffer used for encapsulate
 * and transmit of Virtual File records.
 */
public class DataExchangeBuffer extends OdetteExchangeBuffer {

	private static final long serialVersionUID = 1L;

	private static final int MIN_SUBRECORD_LENGTH = 2;

	private BufferMappingStrategy bufferFilling;

	public DataExchangeBuffer(ByteBuffer rawBuffer) throws OdetteFTPException {
		super();
		parse(rawBuffer);
	}

	public DataExchangeBuffer(int exchangeBufferSize) {
		super();

		buffer = ByteBuffer.allocateDirect(exchangeBufferSize);
		clear();
	}

	public void clear() {
		buffer.clear();
		setIdentifier(CommandIdentifier.DATA);
		buffer.position(1);
		setLength(0);
	}

	protected void parse(ByteBuffer rawBuffer) throws OdetteFTPException {
		int remaining;

		if (rawBuffer == null) {
			throw new IllegalArgumentException("Exchange Buffer is null");
		}

		remaining = rawBuffer.remaining();
		if (remaining == 0) {
			throw new BufferCorruptedException(
					"invalid remaining buffer size: " + remaining);

		}

		setBuffer(rawBuffer.slice());
		setLength(remaining);
	}

	private BufferMappingStrategy getBufferMapping(boolean compression) {
		if (bufferFilling == null) {
			bufferFilling = factoryBufferMappingStrategy(compression);
		}

		return bufferFilling;
	}

	private BufferMappingStrategy factoryBufferMappingStrategy(
			boolean compression) {

		BufferMappingStrategy obj = null;

		if (compression) {
			obj = new CompressionMapping(this);
		} else {
			obj = new NormalMapping(this);
		}

		return obj;
	}

	public int fill(ByteBuffer record, boolean compression,
			RecordFormat format, boolean lastRecord) throws OdetteFTPException {

		BufferMappingStrategy strategy = getBufferMapping(compression);
		return strategy.fill(record, format, lastRecord);
	}

	public ByteBuffer scrape(boolean compression, RecordFormat format)
			throws OdetteFTPException {

		BufferMappingStrategy strategy = getBufferMapping(compression);
		return strategy.scrape(compression, format);
	}

	// public ByteBuffer getRealBuffer() throws OdetteFTPException {
	//
	// int realBufferSize = 0;
	// List<ByteBuffer> bufferList = new LinkedList<ByteBuffer>();
	//
	// /* run all subrecords getting real data and count total size */
	// Iterator<Subrecord> subIt = getSubrecordIterator();
	// while (subIt.hasNext()) {
	// Subrecord sub = subIt.next();
	// ByteBuffer realData = sub.getRealData();
	//
	// realBufferSize += realData.capacity();
	// bufferList.add(realData);
	// }
	//
	// /* allocate the real buffer to be filled and returned */
	// ByteBuffer realBuffer = ByteBuffer.allocateDirect(realBufferSize);
	//
	// Iterator<ByteBuffer> bufIt = bufferList.iterator();
	// while (bufIt.hasNext()) {
	// ByteBuffer realData = bufIt.next();
	// realBuffer.put(realData);
	// }
	//
	// realBuffer.rewind();
	// return realBuffer;
	// }

	public boolean hasRemaining() {
		return (buffer.remaining() >= MIN_SUBRECORD_LENGTH);
	}

	public void prepareSubmit() {
		buffer.flip();
		setLength(buffer.limit());
	}

}
