/**
 *  OdetteJ - Open Source Odette FTP implementation in Java(TM).
 *  Copyright (C) 2005 Fossil E-Commerce, http://www.fossilec.com
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  ------------------------------------------------------------------------
 *  $Id: StreamTransmissionService.java,v 1.3 2005/09/05 13:36:36 rmarins Exp $
 *  ------------------------------------------------------------------------
 */

package org.fossilec.odettej.service;

import java.nio.ByteBuffer;

import org.fossilec.odettej.OdetteFTPException;
import org.fossilec.odettej.transport.NotConnectedException;
import org.fossilec.odettej.transport.Transport;

/**
 * The <code>Service</code> class provide a protocol level abstraction layer
 * which encloses the transport layer in using communication. It's used in
 * conjunction with <code>OdetteExchangeBuffer</code> class to send and
 * receive commands and data.
 * <p>
 * The Odette FTP specification don't indicate whether this service layer must
 * exist or not, but it makes sense to link the Odette Entity and the (network)
 * transport layer.
 * 
 * @author Rafael Marins
 */
public class StreamTransmissionService extends Service {

	public StreamTransmissionService(Transport transport, int debSize) {
		super(transport, debSize);
	}

	public OdetteExchangeBuffer receive(boolean strictFormat)
			throws OdetteFTPException {

		/*
		 * This implementation may face problems in receiving data without some
		 * kind of timeout control. In case of being implemented, consider
		 * handling timeout regarding download of large files which may exceed
		 * timeout even when receiving it.
		 */

		if (!isConnected()) {
			throw new NotConnectedException();
		}

		/* receive and adjust limit to the received data */
		ByteBuffer buffer = receiveStreamTransmissionBuffer();

		/* encapsulate raw buffer into a Stream Transmission Buffer object */
		StreamTransmissionBuffer stb = new StreamTransmissionBuffer(buffer,
				strictFormat);

		return stb.getExchangeBuffer();
	}

	private ByteBuffer receiveStreamTransmissionBuffer()
			throws OdetteFTPException {

		Transport transport = getTransport();

		ByteBuffer sth = ByteBuffer
				.allocateDirect(StreamTransmissionBuffer.HEADER_LENGTH);
		transport.receive(sth);

		int blength = BufferUtil.parseHeaderLengthField(sth);

		ByteBuffer stb = ByteBuffer.allocateDirect(blength);

		sth.rewind();
		stb.put(sth);

		while (stb.hasRemaining()) {
			transport.receive(stb);
		}

		stb.rewind();
		return stb;
	}

	public int send(OdetteExchangeBuffer oeb) throws OdetteFTPException {

		Transport transport = getTransport();
		StreamTransmissionBuffer stb = new StreamTransmissionBuffer(oeb);

		int octetsSent = transport.send(stb.getBuffers());

		return (octetsSent - StreamTransmissionBuffer.HEADER_LENGTH);
	}

}