/**
 *  OdetteJ - Open Source Odette FTP implementation in Java(TM).
 *  Copyright (C) 2005 Fossil E-Commerce, http://www.fossilec.com
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  ------------------------------------------------------------------------
 *  $Id: InternetAddress.java,v 1.4 2005/08/15 16:44:46 rmarins Exp $
 *  ------------------------------------------------------------------------
 */

package org.fossilec.odettej.transport.tcp;

import java.net.InetSocketAddress;

import org.fossilec.odettej.transport.AddressException;
import org.fossilec.odettej.transport.NetworkAddress;

/**
 * This class represents the TCP/IP network address used over the API with
 * Internet transport layer implementation. The address type is
 * <code>internet</code>.
 * <p>
 * In general form it's formed with Hostname (or IP address) and the Port
 * number.
 * 
 * @author Rafael Marins
 * @see org.fossilec.odettej.transport.NetworkAddress
 */
public class InternetAddress extends NetworkAddress {

	/**
	 * Serial Version UID.
	 */
	private static final long serialVersionUID = 1L;

	/**
	 * Internet address type (<code>internet</code>).
	 */
	private static final String ADDRESS_TYPE = "internet";

	/**
	 * Stores address information.
	 */
	private InetSocketAddress location;

	/**
	 * Common way of creating an Address object using the String representation
	 * of the address.
	 * 
	 * @param address
	 *            String representation of the network address.
	 * @throws AddressException
	 *             Fail to parse the address.
	 * @see NetworkAddress#Address(String)
	 */
	public InternetAddress(String address) throws AddressException {
		/* the superclass invoke the parse method */
		super(address);
	}

	/**
	 * Way of creating the Address object using properly parameters.
	 * 
	 * @param host
	 *            <code>String</code> hostname.
	 * @param port
	 *            <code>int</code> port number.
	 * @throws AddressException
	 *             If fail to create the address.
	 */
	public InternetAddress(String host, int port) throws AddressException {
		super();
		this.location = createAddress(host, port);
	}

	/**
	 * Gets the hostname part of the address.
	 * 
	 * @return String hostname.
	 */
	public final String getHost() {
		return location.getHostName();
	}

	/**
	 * Return the port number, a non-negative integer.
	 * 
	 * @return Port number, or <code>0</code> if undefined.
	 */
	public final int getPort() {
		return location.getPort();
	}

	/**
	 * Gets address type.
	 * 
	 * @return String address type "<code>internet</code>".
	 */
	public String getType() {
		return ADDRESS_TYPE;
	}

	/**
	 * Return the String representation of this address object.
	 * 
	 * @return String representation of this address.
	 */
	public String toString() {
		return location.toString();
	}

	/**
	 * Allow the setting up of subclasses context when invoked created through
	 * general constructor <code>NetworkAddress(String)</code>.
	 * 
	 * @param address
	 *            String representation of the network address.
	 * @throws AddressException
	 *             Fail to parse the address.
	 * @see NetworkAddress#parse(String)
	 */
	protected void parse(String address) throws AddressException {
		location = createAddress(address);
	}

	/**
	 * 
	 * @param address
	 * @return
	 * @throws IllegalArgumentException
	 *             If the port parameter is outside the range of valid port
	 *             values, of if the hostname parameter is <code>null</code>.
	 * @throws AddressException
	 *             If failt to create the address.
	 */
	protected static InetSocketAddress createAddress(String address)
			throws AddressException {

		String host = null;
		int port = 0;

		if (address == null) {
			throw new IllegalArgumentException("Address must not be null.");
		} else if (stringContains(address, ":")) {
			String[] parts = address.split(":", 2);
			host = parts[0];
			port = Integer.parseInt(parts[1]);
		} else {
			host = address;
		}

		return createAddress(host, port);
	}

	private static boolean stringContains(String text, String value) {
		return (text.indexOf(value) != -1);
	}

	/**
	 * 
	 * @param host
	 * @param port
	 * @return
	 * @throws IllegalArgumentException
	 *             If the port parameter is outside the range of valid port
	 *             values, of if the hostname parameter is <code>null</code>.
	 * @throws AddressException
	 *             If fail to create the address.
	 */
	protected static InetSocketAddress createAddress(String host, int port)
			throws AddressException {

		InetSocketAddress addr = new InetSocketAddress(host, port);
		return addr;
	}
}
